# process csv files
import time
import datetime
#import cv2
import glob
import os
import pandas as pd
import math

import numpy as np
#from pylab import *
from hilbertcurve.hilbertcurve import HilbertCurve
import random 

from PIL import Image

from sklearn.cluster import AgglomerativeClustering
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.spatial.distance import squareform
import scipy.cluster.hierarchy as sch

from scipy.stats import pearsonr
from scipy.stats import spearmanr
from scipy.stats import kendalltau
from scipy.stats import normaltest

random.seed(0)


def coord_distance(lat1,lon1,lat2,lon2):
    R = 6373.0 #radius of the Earth
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    #Haversine formula
    a = math.sin(dlat / 2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2)**2

    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    distance = R * c
    return distance


def calculate_distM(latlon):
    # convert colums lat and lon to array and then radians

    nelems,x=np.shape(latlon)
    # init distance matrix
    distM = np.zeros([nelems,nelems])
    # lat [x,0] lon [x,1]
    for i in range(nelems-1):
        for j in range(i+1,nelems):
            distM[i,j]=coord_distance(latlon[i,0],latlon[i,1],latlon[j,0],latlon[j,1])
            distM[j,i]=distM[i,j]
            
    return distM,nelems



def highestPowerof2(n): 
    i = n
    while ((i & (i - 1)) != 0):  
        i=i+1
    return i


def plot_distances(idx, dM, cx, cy):
    nelems,x=np.shape(dM)
    x=np.zeros(nelems**2)
    y=np.zeros(nelems**2)
    cont=0

    for conti, i in enumerate(idx):
        for contj, j in enumerate(idx):
            x[cont] = dM[i,j]
            y[cont] = np.linalg.norm(np.array([cx[conti]-cx[contj],cy[conti]-cy[contj]]))
            cont+=1

    corrp,pp = pearsonr(x, y)
    corrs,ps = spearmanr(x, y)
    tau, p_value = kendalltau(x, y)
    k2, px = normaltest(x)
    k2, py = normaltest(y)
    print("pearson correlation {} \nspearman correlation {}\n kendall {}\n".format(corrp,corrs,tau))
    #print("p val for x {} and p val for y {}".format(px,py))
       

def compute_Hilbert_position(latlon, distM, nelems):      
    print("compute hillbert curve for the nodes and their coordinates")
    p = int(max([np.ceil(np.log2(nelems)),1]))
    N = 2
    hilbert_curve = HilbertCurve(p, N)
    mx=highestPowerof2(int(np.ceil(np.sqrt(2**p))))

    np.fill_diagonal(distM, 0)

    mediandist = np.median(distM)
    vdM = squareform(distM)

    Z = linkage(vdM, 'ward',optimal_ordering=True)
    clusters = sch.fcluster(Z, 2.1*mediandist, 'distance')
    #clusters = sch.fcluster(Z, 2.0, 'distance')
    print(np.max(clusters))

    nodes = np.argsort(clusters)

    tempA=distM[nodes,:]
    tempA=tempA[:,nodes]

    cx=np.zeros(nelems)
    cy=np.zeros(nelems)
    cont = 0
    prevc = 1

    for i in range(nelems-1):
        n2 = nodes[i]
        clust = clusters[n2]

        coo=hilbert_curve.point_from_distance(cont)
        if prevc == clust:
            cont += 1  # add here the jumps
        else:
            cont += 4
            prevc = clust
        cx[n2]=coo[0]
        cy[n2]=coo[1]
        #print("n2 {} cx {} cy {}".format(n2,cx[n2],cy[n2]))
    
    plot_distances(nodes, tempA, cx, cy)

    return(nodes,cx,cy,mx)


def make_images(kwh, nodes, cx, cy, mx, path, fname_im):

    #print(path+'hilbert/'+fname_im+'.png')
    #return
    nelems = nodes.size
    RV = [x for x in range(0, nelems)]
    np.random.shuffle(RV)

    _MAXkwh = 5.5
    mxValsH=np.zeros((mx,mx)) # rep with Hilbert curve
    mxValsS=np.zeros((mx,mx)) # rep with seq locality
    mxValsR=np.zeros((mx,mx)) # rep random

    for i, n2 in kwh.items():
        # sequential localityrepresentation
        ux,uy=np.unravel_index(int(nodes[i]), (mx,mx))
        mxValsS[ux,uy]=min(1.0,n2/_MAXkwh)

        # random
        ux,uy=np.unravel_index(RV[i], (mx,mx))
        mxValsR[ux,uy]=min(1.0,n2/_MAXkwh)

        # hilbert
        mxValsH[int(cx[i]),int(cy[i])]=min(1.0,n2/_MAXkwh)
        

                
    imgR = Image.fromarray(mxValsR, 'P')
    imgS = Image.fromarray(mxValsS, 'P')
    imgH = Image.fromarray(mxValsH, 'P')
    
    # test what is the difference between the array and what was written in the image
    
    #Image.fromarray(temp).save('test.tif') 

    imgR.save(path+'random/'+fname_im+'.tif')
    #np.save(path+'random/'+fname_im+'.npy', mxValsR)

    imgS.save(path+'sequential/'+fname_im+'.tif')
    #np.save(path+'sequential/'+fname_im+'.npy', mxValsS)

    imgH.save(path+'hilbert/'+fname_im+'.tif')
    #np.save(path+'hilbert/'+fname_im+'.npy', mxValsH)


    
def save_relevant_quantities(csvfile,kwh,cname):
    with open(csvfile, 'a') as fp:
        fp.write("{},{},{},{},{},{}\n".format(cname, kwh.sum(), kwh.max(), kwh.min(), kwh.mean(), kwh.median()))
    


    
    
# parse all files ---------------------------------------------

path = "rawcsv_data/"
path2= "img_pred_new/"
fname_out = 'comp_71.npz'
                
cont=0
flag = False  # compute placement: True, use file: False

colnames=[]
for file in sorted(glob.glob(path+"*.csv")):
    fname=os.path.basename(file)
    tms=int(time.mktime(datetime.datetime.strptime(fname, "%d-%b-%Y %H:%M:%S.csv").timetuple()))
    cname=str(tms)
    colnames.append(cname)
    print("{} -> {}".format(fname,tms))
    if flag and cont==0:
        df1 = pd.read_csv(file, usecols=[2,3,5], names=['latd','longd',cname])
        cont+=1
        #df1=df1.sort_values(by=['id'])
        df1=df1.sort_values(['longd','latd'])

        kwh = df1[cname]
                
        latlon=df1[{'latd','longd'}].to_numpy()
        latlon=np.radians(latlon)
        # compute radial distance between lat lon coordinates
        distM, nelems = calculate_distM(latlon)
        # determine a partial ordering of the lat lon coordinates
        # use the partial ordering to computer Hilbert coordinates
        nodes, cx, cy, mx = compute_Hilbert_position(latlon, distM, nelems)

        np.savez(fname_out, nodes=nodes, cx=cx, cy=cy, mx=mx)

        break

    else:
        if not flag:
            data = np.load(fname_out)
            nodes = data['nodes']
            cx = data['cx']
            cy = data['cy']
            mx = data['mx']

        df2 = pd.read_csv(file, usecols=[2,3,5], names=['latd','longd',cname])
        df2 = df2.sort_values(['longd','latd'])
        kwh = df2[cname]

    
    save_relevant_quantities(path2+'log.csv',kwh,cname) 
                
    make_images(kwh, nodes, cx, cy, mx, path2, cname)

